open Syntax
open Domain

let opt = ref false
and verbose = ref 0

(* Association variables -> cells *)

let t = Hashtbl.create 23

let print_env chan =
  Hashtbl.iter
    (fun x rd ->
      let d = !rd in
      if d <> Domain.universe then
        Printf.fprintf chan "%s:%a " x Domain.pretty d) t

let assoc_cell x =
  try
    Hashtbl.find t x
  with
  | Not_found ->
      let r = ref Domain.universe in
      Hashtbl.add t x r ;
      r

let c_equals x d kt kf =
  let cell = assoc_cell x in
  fun  env ->
    let v = !cell in
    let vt = Domain.inter v d in
    if not (is_empty vt) then begin
      cell := vt ;
      kt env
    end ;
    let vf = Domain.inter v (complement d) in
    if not (is_empty vf) then begin
      cell := vf ;
      kf env
    end ;
    cell := v

let c_equal x i kt kf =
  let cell = assoc_cell x in
  fun  env ->
    let v = !cell in
    let vt = Domain.inter_elem v i in
    if not (is_empty vt) then begin
      cell := vt ;
      kt env
    end ;

    let vf = Domain.inter_coelem v i in
    if not (is_empty vf) then begin
      cell := vf ;
      kf env
    end ;
    cell := v


(* record variable counts *)

let dump_count chan t =
  if !verbose > 0 then begin
    let l =
      Hashtbl.fold (fun k x r -> (k,x)::r) t [] in
    let l =
      List.sort (fun (_,i1) (_,i2) -> i2-i1) l in
    Printf.fprintf chan "nvars=%d\n" (List.length l) ;
  end

let add_count t x =
  try
    Hashtbl.replace t x (Hashtbl.find t x + 1)
  with
  | Not_found ->
      Hashtbl.add t x 1

let get_count t x =
  try Hashtbl.find t x with
  | Not_found -> assert false

let rec c_count t = function
  | Equals (x,_) -> add_count t x
  | Or es|And es ->
      List.iter (c_count t) es

and s_count t = function
  | Decision (_,_) -> ()
  | If (c, st, [], sf) ->
      c_count t c ; s_count t st ; s_count t sf
  | If (c, st, Elseif (cf,sf) :: el, sd) ->
      c_count t c ; s_count t st ; s_count t (If (cf, sf, el, sd))
  | Case (x, [], d) ->
      s_count t d
  |  Case (x, Arm (is, st)::rem, d) ->
      s_count t st ; s_count t (Case (x, rem, d))

and rs_count rs =
  let t = Hashtbl.create 17 in
  let rec c_rec t = function
    | [] -> t
    | Rule (_,s)::rem ->
        s_count t s ; c_rec t rem in
  c_rec t rs
        

  
(* Compile character to enumeration fonction *)

let rec e_var = function
  | Equals (x,_) -> x
  | Or (e::_)|And (e::_) -> e_var e
  | _ -> ""

let rec e_size = function
  | Equals (_,_) -> 1
  | Or es|And es ->
      List.fold_left (fun r e -> r+e_size e) 0 es

let rec sort_es t es =
  let l1 = List.map
      (fun e ->
        let e = sort_e t e in
        e,e_size e,e_var e) es in
  let l2 = List.sort
      (fun (_,i1,x1) (_,i2,x2) ->
        if i1 < i2 then -1
        else if i1 > i2 then 1
        else Pervasives.compare (get_count t x2) (get_count t x1)) l1 in
  List.map (fun (e,_,_) -> e) l2

and sort_e t = function
  | Or es -> Or (sort_es t es)
  | And es -> And (sort_es t es)
  | e -> e

let c_or e1 e2 kt kf =  e1 kt (e2 kt kf)
and c_and e1 e2 kt kf = e1 (e2 kt kf) kf
and c_false kt kf = kf
and c_true kt kf = kt


type result = (new_state * string) option

(* Compile a condition to an eval fonction *)

let c_equals_approx_true x i kt kf =
  let r = assoc_cell x in
  fun arg -> match Domain.decide i !r with
  | Oui -> kt arg
  | _   -> kf arg

and c_equals_approx_false x i kt kf =
  let r = assoc_cell x in
  fun arg -> match Domain.decide  i !r with
  | Non -> kt arg
  | _   -> kf arg
    

let rec c_approx_true e kt kf = match e with
| Equals (x,i) -> c_equals_approx_true x i kt kf
| Or []        -> kf
| Or (e::es)        ->
    c_or (c_approx_true e) (c_approx_true (Or es)) kt kf
| And []       -> kt
| And (e::es)  ->
    c_and (c_approx_true e) (c_approx_true (And es)) kt kf

let rec c_approx_false e kt kf = match e with
| Equals (x,i) -> c_equals_approx_false x i kt kf
| Or []        -> kt
| Or (e::es)        ->
    c_and (c_approx_false e) (c_approx_false (Or es)) kt kf
| And []       -> kf
| And (e::es)  ->
    c_or (c_approx_false e) (c_approx_false (And es)) kt kf

let rec c_condition e kt kf = match e with
| Equals (x, i) -> c_equal x i kt kf
| Or []   -> kf
| Or (e::es) ->
    let enum_f = c_condition (Or es) kt kf in
    let enum = c_condition e kt enum_f in
    if !opt then
      c_approx_true e kt (c_approx_false e enum_f enum)
    else
      enum
| And []  -> kt
| And (e::es) ->
    let enum_t = c_condition (And es) kt kf in
    let enum = c_condition e enum_t kf in
    if !opt then
      c_approx_true e enum_t (c_approx_false e kf enum)
    else
      enum

let rec c_stm tc s k = match s with
| If (c, st, [], sf) ->
    c_if tc c st sf k
| If (c, st, Elseif (cf,sf) :: el, sd) ->
    c_if tc c st (If (cf, sf, el, sd)) k
| Decision (ns, utter) -> k (Some (ns,utter))
| Case (x, [], d) -> c_stm tc d k
| Case (x, Arm (is, st)::rem, d) ->
    c_equals x
      (Domain.of_ints is) (c_stm tc st k) (c_stm tc (Case (x, rem, d)) k)

and c_if tc c st sf k =
  let c = if !opt then sort_e tc c else c in
  let enum_t = c_stm tc st k and enum_f = c_stm tc sf k in
  let enum = c_condition c enum_t enum_f in
  if !opt then
    c_approx_true c enum_t (c_approx_false c enum_f enum)
  else
    enum

let c_rules rs k = 
  let tc = rs_count rs in
  let rec c_rec rs k = match rs with
  | [] -> k None
  | Rule (is, s)::rem ->
      c_equals "state" (Domain.of_ints is)
        (c_stm tc s k)
        (c_rec rem k) in
  let r = c_rec rs k in
  if !verbose > 0 then dump_count stderr tc ;
  r

let state = assoc_cell "state"

open Printf

let pres chan d = match d with
| None -> fprintf chan "No match"
| Some (None, utter) ->
    fprintf chan "%a %s" Domain.pretty !state utter
| Some (Some i, utter) ->
    fprintf chan "-> %d %s" i utter

let pretty chan d =
  print_env chan ;
  output_string chan " -> " ;
  fprintf chan "%a\n" pres  d
;;

let eval rs =
  c_rules rs
    (fun d _ -> pretty stdout d)
    ()
;;

let incr_verbose d r =
  incr r ;
  if !r mod !verbose = 0 then begin
    print_env stderr ;
    pres stderr d ;
    prerr_endline "" 
  end

let incr_normal _ r = incr r

let count rs =
  if !verbose > 0 then begin
    prerr_string "Compiling... " ;
    flush stderr ;
  end ;
  let incr = if !verbose > 0 then incr_verbose else incr_normal in
  let r = ref 0 in
  let crs =  c_rules rs  (fun d _ -> incr d r) in
  crs () ;
  print_int !r ; print_endline ""
;;

open Printf 

exception Exit

let pstate chan r1 r2 =
  print_env chan ;
  fprintf chan "\n%a\n%a\n" pres r1 pres r2 ;
  flush chan
  
let state_is i = Domain.equal (singleton i) !state

let equiv rs1 rs2 =
  let c1 = ref 0 and c2 = ref 0 in
  let r1 = ref None in

  let k_final r2 _ =
    incr c2 ;
    if !verbose > 0 && !c2 mod !verbose = 0 then begin
      Printf.fprintf stderr "c1=%d, c2=%d\n" !c1 !c2 ;
      pstate stderr !r1 r2
      end ;
    match !r1, r2 with
  | None, None -> ()
  | Some (None, u1), Some (None, u2) when u1=u2 -> ()
  | Some (Some i1, u1), Some (None, u2) when u1=u2 && state_is i1 -> ()
  | Some (None, u1), Some (Some i2, u2) when u1=u2 && state_is i2 -> ()
  | Some (Some i1, u1), Some (Some i2, u2) when u1=u2 && i1=i2 -> ()
  | r1, r2 ->
      print_string "Mismatch: " ;
      pstate stdout r1 r2 ;
      raise Exit in

  if !verbose > 0 then prerr_endline "Compile second" ;
  let c_rs2 = c_rules rs2 k_final in

  if !verbose > 0 then prerr_endline "Compile first" ;
  let c_rs1 =  c_rules rs1  (fun d x -> incr c1 ; r1 := d ;  c_rs2 x) in

  begin try c_rs1 () with Exit -> () end ;
  if !verbose > 0 then
    Printf.fprintf stderr "c1=%d, c2=%d\n" !c1 !c2

    
    
